var lowercaseFirstLetter = function (string) {
    return string.charAt(0).toLowerCase() + string.slice(1);
}

var ChildCollection = function (bridge, sceneComponent, initial) {
    var changedChildren = [];

    var self = new Array();
    return Object.assign(self, {
        initial: initial,
        sceneComponent: sceneComponent,
        add: function (child, skip) {
            this.push(child);

            if (skip) return;

            changedChildren.push(child);
            bridge.UpdateChildren(sceneComponent.entity, changedChildren);
            changedChildren = [];
        },
        remove: function (child) {
            var index = this.indexOf(child);
            if (index != -1) {
                this.splice(index, 1);
            }

            changedChildren.push(child);
            bridge.RemoveChildren(sceneComponent.entity, changedChildren);
            changedChildren = [];
        },
        clear: function () {
            this.length = 0;
        }
    });
};

var ComponentHandler = {
    addComponent: function (runtime, entity, component) {
        if (!component.isTyped) throw "Component is not typed.";
        if (component.entity != undefined) throw "Component may not be added to multiple entities.";

        var typeId = runtime.bridge.getTypeId(component.name);
        if (typeId == Constants.NoType) throw "Component type invalid.";

        runtime.bridge.AddComponent(entity, typeId);
        component.entity = entity;

        ComponentHandler.reflectComponent(runtime, component, component.name);

        return component;
    },
    wrapComponentFactory: function (runtime, name) {
        return function (container) {
            var component = Component.wrap(container, name);

            ComponentHandler.reflectComponent(runtime, component, name);

            return component;
        };
    },
    createComponentFactory: function (runtime, name) {
        return function () {
            var component = Component.create(name);

            component.isTyped = true;

            return component;
        };
    },
    wrapComponent: function (runtime, entity, name) {
        if (!runtime[name]) {
            runtime[name] = {
                wrap: ComponentHandler.wrapComponentFactory(runtime, name),
                create: ComponentHandler.createComponentFactory(runtime, name),
            };
        }

        return runtime[name].wrap(entity, name);
    },
    initializeComponents: function (runtime, extensions) {
        var componentList = [
            "MeshComponent",
            "MeshRenderComponent",
            "MaterialComponent",
            "TransformComponent",
            "MeshGeneratorComponent",
            "MaterialComponent",
            "SceneComponent",
            "EnvironmentComponent",
            "VolDensityFieldComponent",
            "SimSalma_SimulatorComponent"
        ];

        //ExtensionRegistry.extensions.forEach(function (extension) {
        //    if (extensions && extensions.indexOf(extension.name) != -1) {
        //        componentList = componentList.concat(extension.components);
        //    }
        //});

        componentList.forEach(function (name) {
            runtime[name] = {
                wrap: ComponentHandler.wrapComponentFactory(runtime, name),
                create: ComponentHandler.createComponentFactory(runtime, name),
            };
        });

        runtime.SceneComponent = {
            wrap: function (container) {
                var sc = Component.wrap(container, "SceneComponent");

                ComponentHandler.reflectComponent(runtime, sc, "SceneComponent");

                sc.children = ChildCollection(runtime.bridge, sc, true);

                return sc;
            },
            create: function () {
                var sc = Component.create("SceneComponent");

                sc.isTyped = true;
                sc.children = ChildCollection(runtime.bridge, sc, true);

                return sc;
            }
        };
    },
    addField: function (runtime, component, fieldName, typeId, fieldTypeId, i) {
        var self = {
            entity: component.entity,
            componentId: typeId,
            value: null,
            fieldId: i,
            setter: runtime.IdMap[fieldTypeId],
            getter: runtime.GetterIdMap[fieldTypeId],
        };

        try {
            Object.defineProperty(component, fieldName, {
                enumerable: true,
                configurable: false,
                get: function () {
                    if (self.getter != undefined) {
                        self.value = self.getter.call(runtime.bridge, self.entity, self.componentId, self.fieldId, self.value);
                    }
                    return self.value;
                },
                set: function (value) {
                    self.value = value; self.setter.call(runtime.bridge, self.entity, self.componentId, self.fieldId, value);
                },
            });
        } catch (e) {
            console.error(e);
        }

        return self;
    },
    updateComponentInfo: function (runtime, componentName) {
        var componentInfo = {
            map: [],
        };

        runtime.ComponentTemplateMap[componentName] = componentInfo;

        var typeId = runtime.bridge.getTypeId(componentName);

        var numFields = runtime.bridge.getNumFields(typeId);

        for (var i = 0; i < numFields; i++) {
            var fieldTypeId = runtime.bridge.getFieldType(typeId, i);
            var fieldName = runtime.bridge.getFieldName(typeId, i);

            if (runtime.IdMap[fieldTypeId]) {
                componentInfo.map.push({
                    componentId: typeId,
                    fieldTypeId: fieldTypeId,
                    fieldId: i,
                    name: fieldName
                });
            } else {
                console.warn("No mapping available for field " + fieldName + ", of type " + fieldTypeId + "(" + runtime.bridge.getTypeName(fieldTypeId) + ")");
            }
        }

        var baseTypeId = runtime.bridge.getBaseTypeId(typeId);

        if (baseTypeId != Constants.NoType) {
            componentInfo.base = ComponentHandler.getComponentInfo(runtime, runtime.bridge.getTypeName(baseTypeId));
        }

        return componentInfo;
    },
    getComponentInfo: function (runtime, componentName) {
        var componentInfo = runtime.ComponentTemplateMap[componentName];

        return componentInfo ? componentInfo : ComponentHandler.updateComponentInfo(runtime, componentName);
    },
    reflectComponent: function (runtime, component, componentName) {
        runtime.ComponentTemplateMap = runtime.ComponentTemplateMap || {};

        var componentInfo = ComponentHandler.getComponentInfo(runtime, componentName);

        var addFields = function (componentInfo) {
            componentInfo.map.forEach(function (kv) {
                ComponentHandler.addField(runtime, component, kv.name, kv.componentId, kv.fieldTypeId, kv.fieldId);
            });

            if (componentInfo.base) {
                addFields(componentInfo.base);
            }
        };

        addFields(componentInfo);
    }
};
