static float epsilon5 = 0.00001;
static float cos45 = 0.70710678118; // cos(45deg)

float calculateLineSizeRatio(float2 vec)
{
  // Ratio based on the line length.
  float len = length(vec);
  float ratioLength = min(len * 10 + 0.5, 1);

  // Ratio based on the line angle in relation to +X and +Y axes.
  float2 vecNormal = normalize(abs(vec));
  float dotProduct = max(vecNormal.x, vecNormal.y);
  float ratioAngle = lerp(sqrt(2.0), 1.0, dotProduct*2.0-1.0);

  return max(ratioLength, ratioAngle);
}

/*!
  Expands incoming line segments to screen-aligned quads and initializes weights for
  the corner vertices.

  Remaining work:
   - More flexibility for line rasterization
    - Texture profile/prefiltered contour
    - Texture patterns/colors
    - Endpoint rounding options
    - Outlines etc.
*/
void geometryFunction(GS_PRIMITIVE_TYPE VertexIn In[2], inout TriangleStream<VertexOut> OutputStream)
{
  VertexOut vertex = (VertexOut)0;

  float4 p0 = In[0].Position;
  float4 p1 = In[1].Position;

#ifdef COGS_REVERSE_DEPTH
  float t0 = -p0.z + p0.w;
  float t1 = -p1.z + p1.w;
#else
  float t0 = p0.z + p0.w;
  float t1 = p1.z + p1.w;
#endif

  if (t0 < 0 && t1 < 0) {
    return;
  } else if (t0 < 0) {
    p0 = lerp(p0, p1, -t0 / (t1 - t0));
  } else if (t1 < 0) {
    p1 = lerp(p0, p1, -t0 / (t1 - t0));
  }

  float4 begin = p0 / p0.w;
  float4 end = p1 / p1.w;

  const float aspect = viewportSize.x / viewportSize.y;

  begin.x *= aspect;
  end.x *= aspect;

  float2 difference = (end - begin).xy;

  float d = length(difference * 0.5 * viewportSize.yy);
  
  float2 direction = normalize(difference);
  float2 normal = float2(-direction.y, direction.x);

#ifdef COGS_MSAA
  float2 size = lineWidth / viewportSize.yy;
#else
  float2 size = 2 * (0.5 * lineWidth + lineFeather) / viewportSize.yy;
#endif

  size *= calculateLineSizeRatio(difference);

  float3 normalOffset = float3(normal * size, 0);
  float3 endOffset = float3(direction * size * 0.5, 0);

  endOffset.x /= aspect;
  normalOffset.x /= aspect;

  vertex.Position = p0 - float4(normalOffset + endOffset, 0) * p0.w;
  vertex.Weights = float4(0, 0, 0, d);
  transferAttributes(In[0], vertex);
  OutputStream.Append(vertex);

  vertex.Position = p1 - float4(normalOffset - endOffset, 0) * p1.w;
  vertex.Weights = float4(0, 0, d, d);
  transferAttributes(In[1], vertex);
  OutputStream.Append(vertex);

  vertex.Position = p0 + float4(normalOffset - endOffset, 0) * p0.w;
  vertex.Weights = float4(1, 0, 0, d);
  transferAttributes(In[0], vertex);
  OutputStream.Append(vertex);
  
  vertex.Position = p1 + float4(normalOffset + endOffset, 0) * p1.w;
  vertex.Weights = float4(1, 0, d, d);
  transferAttributes(In[1], vertex);
  OutputStream.Append(vertex);
}

float clipNear(inout float4 p0, in float4 p1)
{
#ifdef COGS_REVERSE_DEPTH
  float t0 = -p0.z + p0.w;
  float t1 = -p1.z + p1.w;
#else
  float t0 = p0.z + p0.w;
  float t1 = p1.z + p1.w;
#endif

  if (t0 < 0 && t1 < 0) {
    return 1;
  } else if (t0 < 0) {
    float f = -t0 / (t1 - t0);
    p0 = lerp(p0, p1, f);
    return f;
  } else {
    return 0;
  }
}

void geometryFunctionAdj(GS_PRIMITIVE_TYPE VertexIn In[4], inout TriangleStream<VertexOut> OutputStream)
{
  const float aspect = viewportSize.x / viewportSize.y;

  VertexOut vertex = (VertexOut)0;

  float4 p0 = In[1].Position;
  float4 p1 = In[2].Position;

  float4 prev = In[0].Position;
  float4 next = In[3].Position;

  float clip0 = clipNear(p0, p1);
  float clip1 = clipNear(p1, p0);

  if (clip0 != 0 && clip1 != 0) return;

  clipNear(prev, p0);
  clipNear(next, p1);

  prev /= prev.w;
  next /= next.w;

  prev.x *= aspect;
  next.x *= aspect;

  float4 begin = p0 / p0.w;
  float4 end = p1 / p1.w;

  begin.x *= aspect;
  end.x *= aspect;

  float2 difference = (end - begin).xy;
  float2 prevDiff = (begin - prev).xy;
  float2 nextDiff = (next - end).xy;

  float d = length(difference * 0.5 * viewportSize.yy);

  float2 direction = normalize(difference);

#ifdef COGS_MSAA
  float size = lineWidth;
#else
  float size = 2 * (0.5 * lineWidth + lineFeather);
#endif

  size *= calculateLineSizeRatio(difference);

  float2 normal0 = float2(-direction.y, direction.x);
  float2 normal1 = float2(-direction.y, direction.x);

  const float miterLimit = 200;
  float e0 = 0;
  float e1 = 0;

  if (length(prevDiff) > epsilon5) {
    float2 prevDir = normalize(prevDiff);
    float2 tangent0 = normalize(prevDir + direction);
    float2 perp0 = float2(-prevDir.y, prevDir.x);
    float2 miter0 = float2(-tangent0.y, tangent0.x);
    float miterLength0 = size / dot(miter0, perp0);

    if (abs(miterLength0) > miterLimit) {
      normal0 *= size;
    } else {
      normal0 = miter0 * miterLength0;
      e0 = 1;
    }
  } else {
    normal0 *= size;
  }
  
  if (length(nextDiff) > epsilon5) {
    float2 nextDir = normalize(nextDiff);
    float2 tangent1 = normalize(direction + nextDir);
    float2 perp1 = float2(-direction.y, direction.x);
    float2 miter1 = float2(-tangent1.y, tangent1.x);
    float miterLength1 = size / dot(miter1, perp1);
    if (abs(miterLength1) > miterLimit) {
      normal1 *= size;
    } else {
      normal1 = miter1 * miterLength1;
      e1 = 1;
    }
  } else {
    normal1 *= size;
  }

  float2 offsetSize = 1 / viewportSize.yy;

  float3 normalOffset0 = float3(normal0 * offsetSize, 0);
  float3 normalOffset1 = float3(normal1 * offsetSize, 0);
  
  float l0 = 0.5 * dot(normalOffset0.xy, direction) * viewportSize.y;
  float l1 = 0.5 * dot(normalOffset1.xy, direction) * viewportSize.y;

  normalOffset0.x /= aspect;
  normalOffset1.x /= aspect;

  vertex.Position = p0 - float4(normalOffset0, 0) * p0.w;
  vertex.Weights = float4(0, e0, 0 - l0, d);
  transferAttributes(In[0], vertex);
  OutputStream.Append(vertex);

  vertex.Position = p0 + float4(normalOffset0, 0) * p0.w;
  vertex.Weights = float4(1, e0, 0 + l0, d);
  transferAttributes(In[0], vertex);
  OutputStream.Append(vertex);

  vertex.Position = p1 - float4(normalOffset1, 0) * p1.w;
  vertex.Weights = float4(0, e1, d - l1, d);
  transferAttributes(In[1], vertex);
  OutputStream.Append(vertex);

  vertex.Position = p1 + float4(normalOffset1, 0) * p1.w;
  vertex.Weights = float4(1, e1, d + l1, d);
  transferAttributes(In[1], vertex);
  OutputStream.Append(vertex);
}
