#include "Engine/TransparencyBlendCommon.hlsl"
#include "Engine/PhongLighting.hlsl"

Texture2D depthTexture;
SamplerState linearSampler;

ByteAddressBuffer               List;
StructuredBuffer<FragmentNode>  Node;
StructuredBuffer<FragmentData>  Data;

PixelOut main(VertexIn In, float4 ScreenPos : SV_Position)
{
  const float backgroundDepth = depthTexture.Load(int3(ScreenPos.xy, 0)).r;

  uint startOffset = 4 * ((uint)ScreenPos.y/HeightDivisor * Width + (uint)ScreenPos.x/WidthDivisor);
  uint rx = List.Load(startOffset);
  uint ix = rx & 0x7fffffffu;

  if (NodeAlloc <= ix) {
    PixelOut BackOut = { float4(0, 0, 0, 1) };
    return BackOut;
  }

  float maxAlpha = 0;
  float maxAlphaDepth = COGS_CLEAR_DEPTH;
  float vanillaDepth = COGS_CLEAR_DEPTH;
  uint maxIx = ~0u;

  uint maxIter = 500;
  uint currentIter = 0;
  while (ix < NodeAlloc) {
    // Driver crash safeguard.
    if (currentIter++ == maxIter) break;

    float depth = Node[ix].Depth;
    FragmentData data = Data[ix & 0x7fffffffu];
    int LightModel = data.Emissive >> 24;
    bool closed = (LightModel & 16) != 0;

    if (closed) {
      float alpha = UnpackColor(data.Color).a;
#ifdef COGS_REVERSE_DEPTH
      if ((maxAlpha < alpha) || ((maxAlpha == alpha) && (depth > maxAlphaDepth)))
#else
      if ((maxAlpha < alpha) || ((maxAlpha == alpha) && (depth < maxAlphaDepth)))
#endif
      {
        maxAlpha = alpha;
        maxAlphaDepth = depth;
        maxIx = ix;
      }
    }
    else {
#ifdef COGS_REVERSE_DEPTH
      vanillaDepth = max(depth, vanillaDepth);
#else
      vanillaDepth = min(depth, vanillaDepth);
#endif
    }
    rx = Node[ix].Next;
    ix = rx & 0x7fffffffu;
  }

  float4 result = float4(0, 0, 0, 1);
  if (maxIx < NodeAlloc) {
    FragmentData data = Data[maxIx & 0x7fffffffu];
    float4 viewSpacePosition = mul(float4(In.ScreenPosition.xy, maxAlphaDepth, 1), inverseProjectionMatrix);
    viewSpacePosition /= viewSpacePosition.w;
    float4 worldSpacePosition = mul(viewSpacePosition, inverseViewMatrix);
    SurfaceLighting lightSurface = (SurfaceLighting)0;
    lightSurface.ViewSpacePosition = viewSpacePosition.xyz;
    unpack(lightSurface.Normal,
           lightSurface.LightModel,
           lightSurface.Color.a,
           lightSurface.Emissive,
           lightSurface.Color.rgb,
           lightSurface.Specular,
           lightSurface.SpecularPower,
           data);

    result = float4(DoLighting(lightSurface, ScreenPos).rgb, 0);
  }

#ifdef COGS_REVERSE_DEPTH
  if (vanillaDepth > maxAlphaDepth) {
#else
  if (vanillaDepth < maxAlphaDepth) {
#endif

    // We have some open surfaces, simple additive blend.
    rx = List.Load(startOffset);
    ix = rx & 0x7fffffffu;
    currentIter = 0;
    while (ix < NodeAlloc) {

      // Driver crash safeguard.
      if (currentIter++ == maxIter) break;

      float depth = Node[ix].Depth;
      FragmentData data = Data[ix & 0x7fffffffu];
      int LightModel = data.Emissive >> 24;
      bool closed = (LightModel & 2) != 0;

#ifdef COGS_REVERSE_DEPTH
      if (!closed && depth > maxAlphaDepth)
#else
      if (!closed && depth < maxAlphaDepth)
#endif
      {

        float4 viewSpacePosition = mul(float4(In.ScreenPosition.xy, maxAlphaDepth, 1), inverseProjectionMatrix);
        viewSpacePosition /= viewSpacePosition.w;
        float4 worldSpacePosition = mul(viewSpacePosition, inverseViewMatrix);
        SurfaceLighting lightSurface = (SurfaceLighting)0;
        lightSurface.ViewSpacePosition = viewSpacePosition.xyz;
        unpack(lightSurface.Normal,
               lightSurface.LightModel,
               lightSurface.Color.a,
               lightSurface.Emissive,
               lightSurface.Color.rgb,
               lightSurface.Specular,
               lightSurface.SpecularPower,
               data);

        result.rgb += lightSurface.Color.a * DoLighting(lightSurface, ScreenPos).rgb;
      }

      rx = Node[ix].Next;
      ix = rx & 0x7fffffffu;
    }
  }

  PixelOut Out = { result };
  return Out;
}
