#version 300 es
precision highp float;
precision highp int;

in vec2 NormalizedCoords;

out vec4 fragColor;

uniform sampler2D depthTexture;

uniform float radius;       // World-space ball from where to draw samples
uniform float distmax;      // Falloff with distance from tangent plane
uniform float overplane;    // Epsilon value to avoid surface acne
uniform float minradius;    // Minimum screenspace radius to sample, in pixels.
uniform float maxradius;    // Maximum screenspace radius to sample, in relative screen size.
uniform float intensity;    // Intensity of AO effect
uniform vec2 viewportSize;
uniform mat4 inverseProjectionMatrix;

vec3 invProj(vec2 t, float d)
{
    vec4 h = inverseProjectionMatrix * vec4(2.0*t - vec2(1), 2.0*d - 1.0, 1.0);
    return (1.0/h.w)*h.xyz;
}

void handleSample(inout float occ,
                  inout float wgt,
                  const in vec2 delta,
                  const in vec3 p,
                  const in vec3 n,
                  const in vec2 rot,
                  const in vec2 tt)
{
    float scale = min(viewportSize.x, viewportSize.y) * min(maxradius, 1.0/abs(p.z));
    scale = max(scale, minradius);

    vec2 t2 = NormalizedCoords + radius * scale * delta * vec2(rot.x * tt.x - rot.y * tt.y,
                                                               rot.y * tt.x + rot.x * tt.y);

    if(any(lessThan(t2, vec2(0.0))) || any(lessThan(vec2(1.0), t2))) return;

    float d = textureLod(depthTexture, t2, 0.0).r;
    if(d == 1.0) return;

    vec3 v = invProj(t2, d) - p;

    float o = clamp(1.0 - dot(n,v), 0.0, 1.0);

    float t0 = max(dot(n, normalize(v))-overplane, 0.0);
    float h = 1.0 - sqrt(max(0.0, 1.0 - t0*t0));


    float t1 = dot(n, v);
    float w = clamp(1.0 - t1/distmax, 0.0, 1.0);

    occ += w*h;
    wgt += w;
}

void main()
{
    vec2 ix = mod(floor(NormalizedCoords * viewportSize), 4.0);

    vec2 delta = 1.0/viewportSize;

    vec2 t00 = NormalizedCoords + vec2(0, 0);
    vec2 t10 = NormalizedCoords + vec2(delta.x, 0);
    vec2 t01 = NormalizedCoords + vec2(0, delta.y);

    float d00 = textureLod(depthTexture, t00, 0.0).r;
    if(d00 == 1.0) {
        fragColor = vec4(0);
        return;
    }
    vec3 p00 = invProj(t00, d00);

    // Find the difference that is most like its neighbour,
    // to avoid flattening a crease by taking a difference over it.

    // Estimate dx

    vec2 tx0 = NormalizedCoords + vec2(-2.0*delta.x, 0);
    vec2 tx1 = NormalizedCoords + vec2(    -delta.x, 0);
    vec2 tx3 = NormalizedCoords + vec2(     delta.x, 0);
    vec2 tx4 = NormalizedCoords + vec2( 2.0*delta.x, 0);
    float dx0 = textureLod(depthTexture, tx0, 0.0).r;
    float dx1 = textureLod(depthTexture, tx1, 0.0).r;
    float dx3 = textureLod(depthTexture, tx3, 0.0).r;
    float dx4 = textureLod(depthTexture, tx4, 0.0).r;
    vec3 px0 = invProj(tx0, dx0);
    vec3 px1 = invProj(tx1, dx1);
    vec3 px3 = invProj(tx3, dx3);
    vec3 px4 = invProj(tx4, dx4);
    float tx_l = abs((px1.z-px0.z) - (p00.z-px1.z));
    float tx_r = abs((px4.z-px3.z) - (px3.z-p00.z));

    vec3 dx = (dx3 == 1.0 || tx_l < tx_r) ? (p00 - px1) : (px3 - p00);

    // Estimate dy

    vec2 ty0 = NormalizedCoords + vec2(0, -2.0*delta.x);
    vec2 ty1 = NormalizedCoords + vec2(0,     -delta.y);
    vec2 ty3 = NormalizedCoords + vec2(0,      delta.y);
    vec2 ty4 = NormalizedCoords + vec2(0,  2.0*delta.y);
    float dy0 = textureLod(depthTexture, ty0, 0.0).r;
    float dy1 = textureLod(depthTexture, ty1, 0.0).r;
    float dy3 = textureLod(depthTexture, ty3, 0.0).r;
    float dy4 = textureLod(depthTexture, ty4, 0.0).r;
    vec3 py0 = invProj(ty0, dy0);
    vec3 py1 = invProj(ty1, dy1);
    vec3 py3 = invProj(ty3, dy3);
    vec3 py4 = invProj(ty4, dy4);
    float ty_l = abs((py1.z-py0.z) - (p00.z-py1.z));
    float ty_r = abs((py4.z-py3.z) - (py3.z-p00.z));
    vec3 dy = (dy3 == 1.0 || ty_l < ty_r) ? (p00 - py1) : (py3 - p00);

    vec3 m = normalize(cross(dx, dy));


    float angle = (2.0*3.14/16.0)*(4.0*ix.y + ix.x);
    vec2 rot = vec2(cos(angle),sin(angle));

    float occ = 0.0;
    float wgt = 0.0;
    handleSample(occ, wgt, delta, p00, m, rot, vec2( 0.1, 0.0));
    handleSample(occ, wgt, delta, p00, m, rot, vec2( 0.0, 0.2));
    handleSample(occ, wgt, delta, p00, m, rot, vec2(-0.3, 0.0));
    handleSample(occ, wgt, delta, p00, m, rot, vec2( 0.0,-0.4));

    float foo = 1.0 - occ * intensity/max(wgt, 0.0001);

    fragColor = vec4(0.5*(m+vec3(1.0)), foo);
}

